/*
 * Copyright (c) 1991-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* This file written by Tom Henderson (tomh@cs.berkeley.edu) */
/* Version 1.0 */

#include <stdlib.h>
#include <math.h>
#include "tclcl.h"
#include "packet.h"
#include "random.h"
#include "ip.h"
#include "stp.h"

static class SDHeaderClass : public PacketHeaderClass {
public:
        SDHeaderClass() : PacketHeaderClass("PacketHeader/SD",
					     sizeof(hdr_sd)) {}
} class_sdhdr;

static class POLLHeaderClass : public PacketHeaderClass {
public:
        POLLHeaderClass() : PacketHeaderClass("PacketHeader/POLL",
					     sizeof(hdr_poll)) {}
} class_pollhdr;

static class STATHeaderClass : public PacketHeaderClass {
public:
        STATHeaderClass() : PacketHeaderClass("PacketHeader/STAT",
					     sizeof(hdr_stat)) {}
} class_stathdr;

static class USTATHeaderClass : public PacketHeaderClass {
public:
        USTATHeaderClass() : PacketHeaderClass("PacketHeader/USTAT",
					     sizeof(hdr_ustat)) {}
} class_ustathdr;

static class STPSinkClass : public TclClass {
public:
	STPSinkClass() : TclClass("Agent/STPSink") {}
	TclObject* create(int argc, const char*const* argv) {
		return (new STPSinkAgent());
	}
} class_stp_sink;

STPSinkAgent::STPSinkAgent() : Agent(PT_STAT), 
	r_highest_(0), r_inseq_(0), dup_(0), closed_(0)
{
	// Defaults for bound variables should be set in ns-default.tcl.
	bind("win_", &win_);
	bind("p_error_", &p_error_);

	finish_[0] = 0;

	// reset used for dynamically created agent
	reset();
}

void STPSinkAgent::reset()
{
	int i;
        for (i = 0; i< B_SIZE; i++) { 
                retrans_buffer[i].ts_ = 0;
                retrans_buffer[i].seqno_ = 0;
                retrans_buffer[i].ack_ = 0;
                retrans_buffer[i].flag_ = 0;
        }

}

void STPSinkAgent::process_sd(Packet* pkt)
{
	hdr_sd* sdh = hdr_sd::access(pkt);
	int seq = sdh->seqno();
	int seq_mod = seq % B_SIZE;
	int i, i_mod;

	double now = Scheduler::instance().clock();
	// Normal path
	if (seq == r_highest_ && r_highest_ == r_inseq_) {
		r_highest_++; 
		r_inseq_++;
		return;	
	} 
	if (seq == r_highest_) { // next in order
		retrans_buffer[seq_mod].ack_ = 1;
		r_highest_++;
		check_for_ustat();
	} else if (seq > r_highest_) { // off top of window
		// clear buffer between r_highest_ and seq
		for (i = r_highest_; i < seq; i++) {
			i_mod = i%B_SIZE;
			retrans_buffer[i_mod].seqno_ = 0;
			retrans_buffer[i_mod].ack_ = 0;
			retrans_buffer[i_mod].flag_ = 0;
		}
		retrans_buffer[seq_mod].ack_ = 1;
		r_highest_ = seq + 1;
		check_for_ustat();
	} else if (seq < r_highest_ && seq >= r_inseq_) { // possible retransmission
		if (retrans_buffer[seq_mod].ack_) { // already got it
			dup_++;
		} else {
			retrans_buffer[seq_mod].ack_ = 1;
		}
		if (seq == r_inseq_) {
			r_inseq_++;
			// Now check if we can advance inseq_ more
			while (r_inseq_ < r_highest_) {  
				i_mod = r_inseq_ % B_SIZE;
				if (retrans_buffer[i_mod].ack_) 
					r_inseq_++;
				else
					break;
			}
			
		}
		check_for_ustat();
	} else if (seq < r_inseq_) {
			dup_++;
	} else {
		printf("Hit a strange case in receiving data\n");
		abort();
	}
	return;
}

void STPSinkAgent::check_for_ustat()
{
	int i, i_mod, j, j_mod;
	int first, second;
	int dupack = 0;
	int found = 0; 
	i = r_highest_ - 1;
	i_mod = i % B_SIZE;
	while (i > r_inseq_ && !found) {
		// find top edge of first gap
		while (retrans_buffer[i_mod].ack_) {	
			dupack++;
			i--;
			i_mod = i % B_SIZE;
		}
		second = i + 1;
		while (!retrans_buffer[i_mod].ack_ && i >= r_inseq_) {
			if (retrans_buffer[i_mod].flag_) {
				found = 1;
				break;
			}
			i--;
			i_mod = i % B_SIZE;
		}
		first = i + 1;
		if (!found && dupack >= NUMDUPACKS) {
			// Send a ustat
		    Packet* p = allocpkt();
			hdr_cmn* th = HDR_CMN(p);
			th->ptype() = PT_USTAT;
			th->size() = 40;

    		hdr_ustat *ustath = hdr_ustat::access(p);
    		ustath->seqno() = r_inseq_;
    		ustath->element_1() = first;
    		ustath->element_2() = second;
			send(p, 0);
			// Mark packets as having been ustat'ed
			for (j = first; j < second; j++) {
				j_mod = j % B_SIZE;
				retrans_buffer[j_mod].flag_ = 1;
			}
		}
	}
}

void STPSinkAgent::process_poll(Packet* pkt)
{
	int i, i_mod;
	hdr_poll* pollh = hdr_poll::access(pkt);
	if (pollh->seqno() > r_highest_) {
        for (i = r_highest_; i < pollh->seqno(); i++) {
            i_mod = i%B_SIZE;
            retrans_buffer[i_mod].seqno_ = 0;
            retrans_buffer[i_mod].ack_ = 0;
            retrans_buffer[i_mod].flag_ = 0;
        }
		r_highest_ = pollh->seqno();
	}
	send_stat(pollh->ts());
}

void STPSinkAgent::send_stat(double tstamp)
{
	int stat_len_ = 0;
	int top, bottom;
	int i, i_mod;
	int working_on_gap, working_on_block;
	working_on_gap = working_on_block = 0;
	Packet* p = allocpkt();
        hdr_cmn* th = HDR_CMN(p);

	hdr_stat *stath = hdr_stat::access(p);
	double now = Scheduler::instance().clock();
	stath->ts() = tstamp;
	stath->seqno() = r_inseq_;
	stath->win() = win_;
	/* add list elements */
	if (r_inseq_ != r_highest_) {
		working_on_gap = 1;  // start off by working on a gap
		top = r_highest_;
		bottom = r_inseq_;
		stath->list_[0] = r_inseq_;
		stat_len_++;
	 	for (i = bottom; i < top; i++) {
			i_mod = i % B_SIZE;
			if (stat_len_ >= MAXSTAT)
				break;
			if (working_on_gap) {
				if (retrans_buffer[i_mod].ack_) {
					stath->list_[stat_len_] = i;
					stat_len_++;
					working_on_block = 1;
					working_on_gap = 0;
				}
			} else { // working_on_block
				if (!retrans_buffer[i_mod].ack_) {
					stath->list_[stat_len_] = i;
					stat_len_++;
					working_on_block = 0;
					working_on_gap = 1;
				}
			}
		} // end of for loop
	}
	stath->list_len() = stat_len_;
    th->size() = 40 + 4*stat_len_;

	send(p, 0);
}

int STPSinkAgent::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		/* a Tcl procedure to be invoked when the connection is done */
		if (strcmp(argv[1], "finish") == 0) {
			strcpy(finish_, argv[2]);
			return (TCL_OK);
		}
	}
	return (Agent::command(argc, argv));
}

void STPSinkAgent::recv(Packet *pkt, Handler*)
{
	hdr_cmn* th = HDR_CMN(pkt);

	double now = Scheduler::instance().clock();
	int pkt_size;
	int errored = 0;
	double prob_error;
	pkt_size = th->size();
	if (p_error_) {
		prob_error = 1 - pow((1-p_error_),(pkt_size * 8));
		double hit = (float)Random::random()/MAX_RAND;
		if (hit < prob_error)
			errored = 1;
	}
	if (!errored) {
		switch(th->ptype()) {
		case PT_SD:
			process_sd(pkt);
			break;

		case PT_POLL:
			process_poll(pkt);
			break;

		default:
			Tcl::instance().evalf("%s error \"received illegal packet type\"",
				      name());
			Packet::free(pkt);
			return;
		}
	} 
	Packet::free(pkt);
}

/*
 * This function is invoked when the connection is done. It in turn
 * invokes the Tcl finish procedure that was registered with STP.
 */
void STPSinkAgent::finish() {
	char wrk[100];

	if (finish_ != "") 
		sprintf(wrk, "%s", finish_);
	else
		sprintf(wrk, "finish", name());
	Tcl::instance().eval(wrk);
}


