/*
 * Copyright (c) 1991-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* This file written by Tom Henderson (tomh@cs.berkeley.edu) */
/* Version 1.0 */

#include <stdlib.h>
#include <math.h>
#include "tclcl.h"
#include "packet.h"
#include "ip.h"
#include "stp.h"

int hdr_sd::offset_;
int hdr_poll::offset_;
int hdr_stat::offset_;
int hdr_ustat::offset_;
static class SDHeaderClass : public PacketHeaderClass {
public:
        SDHeaderClass() : PacketHeaderClass("PacketHeader/SD",
					     sizeof(hdr_sd)) {}
} class_sdhdr;

static class POLLHeaderClass : public PacketHeaderClass {
public:
        POLLHeaderClass() : PacketHeaderClass("PacketHeader/POLL",
					     sizeof(hdr_poll)) {}
} class_pollhdr;

static class STATHeaderClass : public PacketHeaderClass {
public:
        STATHeaderClass() : PacketHeaderClass("PacketHeader/STAT",
					     sizeof(hdr_stat)) {}
} class_stathdr;

static class USTATHeaderClass : public PacketHeaderClass {
public:
        USTATHeaderClass() : PacketHeaderClass("PacketHeader/USTAT",
					     sizeof(hdr_ustat)) {}
} class_ustathdr;

static class STPClass : public TclClass {
public:
	STPClass() : TclClass("Agent/STP") {}
	TclObject* create(int argc, const char*const* argv) {
		return (new STPAgent());
	}
} class_stp;

STPAgent::STPAgent() : Agent(PT_SD), 
	t_seqno_(0), highest_ack_(0), recover_(0), cwnd_(0),
	ssthresh_(0), srtt_(0), rttvar_(0), send_stack_(0), num_to_send_(0),
	curseq_(0), closed_(0), poll_timer_(this), del_send_timer_(this)
{
	// Defaults for bound variables should be set in ns-default.tcl.
	bind("window_", &wnd_);
	bind("windowInit_", &wnd_init_);
	bind("windowOption_", &wnd_option_);
	bind("windowConstant_", &wnd_const_);
	bind("packetSize_", &size_);
	bind("maxcwnd_", &maxcwnd_);
	bind("highest_ack_", &highest_ack_);
	bind("cwnd_", &cwnd_);
	bind("poll_interval_", &poll_interval_);
	bind("initial_rtt_", &initial_rtt_);
	bind("rate_ctl_", &rate_ctl_);
	bind("rate_", &rate_);

	finish_[0] = 0;

	// reset used for dynamically created agent
	reset();
}

void STPAgent::reset()
{
	rtt_init();
	int i;

	curseq_ = 0;
	cwnd_ = wnd_init_;
	t_seqno_ = 0;
	highest_ack_ = 0;
	recover_ = 0;
	ssthresh_ = int(wnd_);
	for (i = 0; i< B_SIZE; i++) {
		trans_buffer[i].ts_ = 0;
		trans_buffer[i].seqno_ = 0;
		trans_buffer[i].ack_ = 0;
		trans_buffer[i].flag_ = 0;
	}
}

/*
 * Initialize variables for the retransmit timer.
 */
void STPAgent::rtt_init()
{
	srtt_ = initial_rtt_;
	rttvar_ = 0.125 * srtt_;
}

/*
 * Rtt updates are similar to those in TCP.   The poll interval is set to
 * poll an integer number of times per RTT.
 */
void STPAgent::rtt_update(double tao)
{
	double delta = tao - srtt_;
	if (delta < 0)
		delta *= -1;
	srtt_ = 0.125 * tao + 0.875 * srtt_;
	rttvar_ = rttvar_ + delta - (rttvar_)/4;

	poll_interval_ = srtt_/POLL_PER_RTT;
}

void STPAgent::send_sd(int seqno)
{
	Packet* p = allocpkt();
	hdr_sd *sdh = hdr_sd::access(p);
	double now = Scheduler::instance().clock();
	sdh->seqno() = seqno;

	send(p, 0);
	/* Store in transmission buffer */
	int index = seqno % B_SIZE;
	trans_buffer[index].ts_ = now;
	trans_buffer[index].seqno_ = seqno;
}

void STPAgent::send_poll()
{
	Packet* p = allocpkt();
	/* Reassign some fields, because type is set to SD by default */
    hdr_cmn* th = HDR_CMN(p);
    th->ptype() = PT_POLL;
    th->size() = 36;

	hdr_poll *pollh = hdr_poll::access(p);
	double now = Scheduler::instance().clock();
	pollh->ts() = now;
	pollh->seqno() = t_seqno_;

	send(p, 0);
}

int STPAgent::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "advance") == 0) {
			curseq_ = atoi(argv[2]);
			send_data();
       		poll_timer_.resched(poll_interval_);
			del_send_timer_.resched(0.001);
			return (TCL_OK);
		}
		/* a Tcl procedure to be invoked when the connection is done */
		if (strcmp(argv[1], "finish") == 0) {
			strcpy(finish_, argv[2]);
			return (TCL_OK);
		}
	}
	return (Agent::command(argc, argv));
}

int STPAgent::window()
{
	return (int(cwnd_ < wnd_ ? cwnd_ : wnd_));
}

int STPAgent::process_stat(Packet* pkt) 
{
	int i, j, j_mod, element1, element1_mod, element2_mod, element2;
	int advance = 0;
	int retransmit = 0;
	hdr_stat* stath = hdr_stat::access(pkt);
	double rtt = Scheduler::instance().clock() - stath->ts();	
	rtt_update(rtt);
	wnd_ = stath->win();
	int stat_seqno = stath->seqno();
	int list_len = stath->list_len();

	advance = stat_seqno - highest_ack_;
	assert((list_len) <= MAXSTAT);  
	if (list_len == 0 && stat_seqno > highest_ack_) {
		advance = stat_seqno - highest_ack_;
		highest_ack_ = stat_seqno;
	}	
	if (list_len > 0) {
		double now = Scheduler::instance().clock();	
		for (i = 0; i < (list_len - 1); i++) {
			element1 = stath->list_[i];
			element2 = stath->list_[i + 1];
			element1_mod = element1 % B_SIZE;
			element2_mod = element2 % B_SIZE;
			for (j = element1; j < element2; j++) {
				j_mod = j % B_SIZE;
				if (now - trans_buffer[j_mod].ts_ > (srtt_ + 2*rttvar_)) { // retransmit
					send_sd(j);
					if (rate_ctl_)
						send_stack_++;
					retransmit = 1;
				}
			}
			if ((i+2) < list_len) {
				element1 = stath->list_[i+2];
				element1_mod = element1 % B_SIZE;
				for (j = element2; j < element1; j++) {
					j_mod = j % B_SIZE;
					trans_buffer[j_mod].ack_ = 1;
				}
			}
			i++;
		} // end of for loop
		if (highest_ack_ > recover_) {
			opencwnd(advance);
			recover_ = t_seqno_;	
			closecwnd(1);
		}
	}
	if (!retransmit && highest_ack_ > recover_ && advance > 0)
		opencwnd(advance);

	return 1;
}

void STPAgent::process_ustat(Packet* pkt)  
{
	// In future, should add sanity checks against missequencing
	int i;
	hdr_ustat* ustath = hdr_ustat::access(pkt);
	highest_ack_ = ustath->seqno();
	int start = ustath->element_1();
	int stop = ustath->element_2();
	for (i=start; i< stop; i++) {
		if (rate_ctl_)
			send_stack_++;
		send_sd(i);
	}
	if (highest_ack_ > recover_) {
		recover_ = t_seqno_;	
		closecwnd(1);
	}
}

/*
 * Try to send as much data as the window will allow.  The link layer will 
 * do the buffering; we ask the application layer for the size of the packets.
 * If rate controlled, do nothing.
 */
void STPAgent::send_data()
{
	if (!rate_ctl_) {
	int win = window();
	int i, i_mod;

	/* see how many new packets can be sent.  Schedule
	 * them to be sent staggered over time
	 * can send new data if pipe_ < cwnd_ and (t_seqno_ - highest_ack_) < wnd_
	 * and t_seqno_ < curseq_
	 */

	// Compute amt of data in the network
	int pipe = t_seqno_ - highest_ack_;   
	if (pipe > 0) {
		double now = Scheduler::instance().clock();
		for (i = highest_ack_; i < t_seqno_; i++) {
			i_mod = i % B_SIZE;
			// If you have sent it long enough ago, can assume it left pipe
			if (trans_buffer[i_mod].ts_ < (now - (srtt_ + 2*rttvar_)))
				pipe--;
 		}
	}
	
	if (send_stack_ > 0) 
		num_to_send_ = send_stack_;
	else
		num_to_send_ = 0;
	while (pipe < cwnd_ && (t_seqno_ - highest_ack_) <= wnd_ && t_seqno_ < curseq_) {
			send_stack_++;
			num_to_send_++;
			pipe++;
	}
	if (num_to_send_)
		del_send_timer_.resched(0.0001);
	} 
}

/*
 * open up the congestion window, based on num acks
 */
void STPAgent::opencwnd(int num)
{
	int i;
	for (i = 0; i< num; i++) {
		if (cwnd_ < ssthresh_) {
			/* slow-start (exponential) */
			cwnd_ += 1;
		} else {
			switch (wnd_option_) {
			case 1:
				/* This is the standard algorithm. */
				cwnd_ += 1 / cwnd_;
				break;

			default:
#ifdef notdef
			/*XXX*/
			error("illegal window option %d", wnd_option_);
#endif
				abort();
			}
		}
	}
	// if maxcwnd_ is set (nonzero), make it the cwnd limit
	if (maxcwnd_ && (int(cwnd_) > maxcwnd_))
		cwnd_ = maxcwnd_;

	return;
}

/*
 * close down the congestion window
 */
void STPAgent::closecwnd(int how)
{   
	double now = Scheduler::instance().clock();
	switch (how) {
	case 0:
		/* timeouts */
		ssthresh_ = int( window() / 2 );
		if (ssthresh_ < 2)
			ssthresh_ = 2; 
		/* Old code: ssthresh_ = int(cwnd_ / 2); */
		cwnd_ = int(wnd_init_);
		break;
        
	case 1:
		/* dup acks, or after a recent congestion indication. */
		cwnd_ = window()/2;
		ssthresh_ = int(cwnd_);
		if (ssthresh_ < 2)
			ssthresh_ = 2;
		if (cwnd_ < 1)
			cwnd_ = 1;
		break;          

	default:
		abort();
	}
}


/*
 * Respond either to a source quench or to a congestion indication bit.
 * This is done at most once a roundtrip time;  after a source quench,
 * another one will not be done until the last packet transmitted before
 * the previous source quench has been ACKed.
 */
void STPAgent::quench(int how)
{
}

/*
 * main reception path - should only see acks, otherwise the
 * network connections are misconfigured
 */
void STPAgent::recv(Packet *pkt, Handler*)
{
	hdr_cmn* th = HDR_CMN(pkt);

	switch(th->ptype()) {
	case PT_STAT:
		process_stat(pkt);
		break;

	case PT_USTAT:
		process_ustat(pkt);
		break;

	default:
		printf("pkt_type = %d\n", th->ptype());
		Tcl::instance().evalf("%s error \"received illegal packet type\"",
				      name());
		Packet::free(pkt);
		return;
	}
	Packet::free(pkt);
	/*
	 * Try to send more new data.
	 */
	send_data();
}

/*
 * This function is invoked when the connection is done. It in turn
 * invokes the Tcl finish procedure that was registered with TCP.
 */
void STPAgent::finish() {
	char wrk[100];

	if (finish_ != "") 
		sprintf(wrk, "%s", finish_);
	else
		sprintf(wrk, "finish", name());
	Tcl::instance().eval(wrk);
}

void PollTimer::expire(Event *e) {
                a_->send_poll();
                a_->poll_timer_.resched(a_->poll_interval_);
}

void DelSendTimer::expire(Event *e) {
	if (!a_->rate_ctl_) {
	int num_left = a_->num_to_send_;
	if (a_->send_stack_ > 0) {
		a_->send_sd(a_->t_seqno_++);
		a_->send_stack_--;
	}
	if (num_left == 0)
		num_left = 1;
	a_->del_send_timer_.resched((a_->srtt_)/num_left);
	} else { // rate control
		if (a_->send_stack_ > 0)
			a_->send_stack_--;
		else
			a_->send_sd(a_->t_seqno_++);
		a_->del_send_timer_.resched(1/a_->rate_);
	}
}

