/*
 * Copyright (c) 1991-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* This file written by Tom Henderson (tomh@cs.berkeley.edu) */
/* Version 1.0 */

#ifndef ns_stp_h
#define ns_stp_h

#include "agent.h"
#include "packet.h"
#include "assert.h"

#define NUMDUPACKS 3		/* normally 3, sometimes 1 */
#define PACKET_POLL 100
#define MAXSTAT 21 /* 10 gaps per stat , should be an odd integer*/
#define POLL_THRESH 8
#define POLL_PER_RTT 3
#define B_SIZE 2048  /* Buffer size */
#define MAX_RAND 2147483648

struct hdr_sd {
	int seqno_;
	int& seqno() {
		return (seqno_);
	}
	// Header access methods
	static int offset_; // required by PacketHeaderManager
	inline static int& offset() { return offset_; }
	inline static hdr_sd* access(const Packet* p) {
		return (hdr_sd*) p->access(offset_);
	}
};

struct hdr_poll {
	double ts_;
	int seqno_;
	double& ts() {
		return (ts_);
	}
	int& seqno() {
		return (seqno_);
	}
	// Header access methods
	static int offset_; // required by PacketHeaderManager
	inline static int& offset() { return offset_; }
	inline static hdr_poll* access(const Packet* p) {
		return (hdr_poll*) p->access(offset_);
	}
};

struct hdr_stat {
	double ts_;
	int win_;
	int list_len_;
	int seqno_;
	int list_[MAXSTAT]; 
	double& ts() {
		return (ts_);
	}
	int& win() {
		return (win_);
	}
	int& list_len() {
		return (list_len_);
	}
	int& seqno() {
		return (seqno_);
	}
	int& list(int num) {
		return (list_[num]);
	}
	// Header access methods
	static int offset_; // required by PacketHeaderManager
	inline static int& offset() { return offset_; }
	inline static hdr_stat* access(const Packet* p) {
		return (hdr_stat*) p->access(offset_);
	}
};

struct hdr_ustat {
	int seqno_;
	int element_1_;
	int element_2_;
	int& seqno() {
		return (seqno_);
	}
	int& element_1() {
		return (element_1_);
	}
	int& element_2() {
		return (element_2_);
	}
	// Header access methods
	static int offset_; // required by PacketHeaderManager
	inline static int& offset() { return offset_; }
	inline static hdr_ustat* access(const Packet* p) {
		return (hdr_ustat*) p->access(offset_);
	}
};

struct segment {
	double ts_;
	int seqno_;	/* Note: may get rid of this */
	int ack_; 	/* flags acks at transmitter, receiver */
	int flag_; 	/* flags whether something has been ustat'ed at rcv */
};
	

class STPAgent;     

class PollTimer : public TimerHandler {
public:
        PollTimer(STPAgent *a) : TimerHandler() { a_ = a; }
protected:
        virtual void expire(Event *e);
        STPAgent *a_;
};  
    
class DelSendTimer : public TimerHandler {
public:
        DelSendTimer(STPAgent *a) : TimerHandler() { a_ = a; }
protected:
        virtual void expire(Event *e);
        STPAgent *a_;
};  
    
class STPAgent : public Agent {
	friend class PollTimer;
	friend class DelSendTimer;
public:
	STPAgent();

	void recv(Packet*, Handler*);
	int command(int argc, const char*const* argv);

protected:
	/*
	 * State encompassing the round-trip-time estimate.
	 */
	double srtt_;     	/* smoothed round-trip time */
	double rttvar_;   	/* variance in round-trip time */
	double initial_rtt_;  	/* presumably obtained via BGN exchange */
	void rtt_init();
	void rtt_update(double tao);

	PollTimer poll_timer_;
	DelSendTimer del_send_timer_;
	double poll_interval_;

	int t_seqno_;		/* sequence number */
	int curseq_;	/* highest seqno produced by app */
	int wnd_;	/* set by receiver */
	double cwnd_;
	int maxcwnd_;		/* max # cwnd can ever be */
	int highest_ack_;	/* highest ack that is in sequence */
	int recover_;		/* used to mark end of fast recovery */
	int ssthresh_;		/* slow start threshold */
	int send_stack_;	/* for delayed sends */
	int num_to_send_;	/* for delayed sends */
	double wnd_const_;
	double wnd_init_;
	int wnd_option_;
	
	int rate_ctl_;
	double rate_;

	segment trans_buffer[B_SIZE];
	
	/* Stats */
	int rtx_;

	int off_ip_;
	int off_sd_;
	int off_poll_;
	int off_stat_;
	int off_ustat_;

	void send_data();
	void send_poll();
	void send_sd(int seqno);
	void opencwnd(int num);
	void closecwnd(int how);
	int process_stat(Packet* pkt);
	void process_ustat(Packet* pkt);
	void quench(int how);
	int window();
	
	void reset();
	void finish(); /* called when the connection is terminated */

	char finish_[20];       /* name of Tcl proc to call at finish time */
	int closed_;            /* whether this connection has closed */
};

class STPSinkAgent : public Agent {
public:
	STPSinkAgent();
	void recv(Packet*, Handler*);
	int command(int argc, const char*const* argv);
protected:
	void send_stat(double tstamp);
	void process_sd(Packet* pkt);
	void process_poll(Packet* pkt);
	void check_for_ustat();

	void reset();
	void finish(); /* called when the connection is terminated */

	char finish_[20];       /* name of Tcl proc to call at finish time */
	int closed_;            /* whether this connection has closed */
	
	int win_;
	int r_highest_;
	int r_inseq_;
	double p_error_;

	segment retrans_buffer[B_SIZE];
	int dup_;

	int off_ip_;
	int off_sd_;
	int off_poll_;
	int off_stat_;
	int off_ustat_;
};

#endif 
